/*****************************************************************************
    icmp.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2003 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef HAVE_IN_SYSTM
#include <netinet/in_systm.h>
#endif
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>

#include "params.h"
#include "globals.h"
#include "logfun.h"
#include "icmp.h"
#include "neterror.h"

#ifndef UDP_DPORT
#error >>> UDP header structure undefined
#endif

#ifndef ICMP_PORT_UNREACH
#ifdef  ICMP_UNREACH_PORT
#define ICMP_PORT_UNREACH ICMP_UNREACH_PORT
#else
#error >>> ICMP Port Unreachable not defined
#endif
#endif


static char *icmpTypes[] = {	"Echo_Reply",
				"1",
				"2",
				"Dest-Unreachable",
				"Source-Quench",
				"Redirect",
				"6",
				"7",
				"Echo",
				"RouterAdvert",
				"Router_Solicit",
				"Time-Exceeded",
				"Parameter-Problem",
				"Timestamp",
				"Timestamp-Reply",
				"Info-Request",
				"Info-Reply",
				"Mask-Request",
				"Mask-Reply" };

static char *icmpUrCodes[] = {	"Net",
				"Host",
				"Protocol",
				"Port",
				"Frag",
				"Source",
				"DestNet",
				"DestHost",
				"Isolated",
				"Auth-Net",
				"Auth-Host",
				"Net-Svc",
				"Host-Svc",
				"Filtered",
				"Precd-Violation",
				"Precd-Cut" };

/*
	convert icmp type/code into text.
	Note: code field decoded for ICMP_UNREACH only !
*/

static char *getIcmpType (uchar typ, uchar cod) 
{
    static char buf[128];

    if (typ < (uchar)(sizeof(icmpTypes)/sizeof(icmpTypes[0]))) {
        if ((typ == ICMP_UNREACH) && 
            (cod < (uchar)(sizeof(icmpUrCodes)/sizeof(icmpUrCodes[0])))) {
            snprintf (buf, sizeof(buf), "%s/%s", 
                      icmpTypes[typ], icmpUrCodes[cod]);
            return buf;
        }
        else
            return icmpTypes[typ];
    }
    return "Unknown";
}


/*
	start a network socket for receiption of icmp messages
*/

int icmp_init()
{
    struct protoent *p;
    int icmp_sock;

    if ((p = getprotobyname("icmp")) == NULL) {
        LOG_MSG1 (LOG_ERR, _("Protocal 'icmp' unknown"));
        return -1;
    }

    if ((icmp_sock = socket(PF_INET, SOCK_RAW, p->p_proto)) < 0) {
        LOG_MSG1 (LOG_ERR, _("No ICMP socket"));
        return -1;
    }

    LOG_MSG5 (LOG_DEBUG, "ICMP socket initialized");
    return icmp_sock;
}


/*
	receive and process an icmp message
*/

int icmp_recv(int icmp_sock, struct sockaddr_in *sp)
{
    struct ip *ipHead, *ip2Head;
    struct icmp *icmp;
    struct udphdr *udpHead;
    int    ipHeadLen, ip2HeadLen, recvLen;
    socklen_t spLen;
    static char recvBuf[1024];

    spLen = sizeof(struct sockaddr_in);

    if ((recvLen = recvfrom(icmp_sock, recvBuf, sizeof(recvBuf), 0,
			    (struct sockaddr *) sp, &spLen)) <= 0) {
        if ((errno == EINTR) || (recvLen == 0))
            return 0;		/* timeout ? ignore this one !	*/
        else {
            logReceiveError("ICMP");
            return -1;
        }
    }

/*
	Get the length of the IP header	from the header itself.
	This avoids trouble if we should get some packet with options appended.
	Immediately following the IP header is the icmp header, which should
	be complete at least.
*/

    ipHead    = (struct ip *)recvBuf;
    ipHeadLen = ipHead->ip_hl << 2;

    if (recvLen <= (ipHeadLen + ICMP_MINLEN)) {
        logShortPacket(recvLen, "ICMP");
        return 0;
    }

   icmp = (struct icmp *)(((char *)ipHead) + ipHeadLen);

    LOG_MSG5 (LOG_DEBUG, "%s type %d:%d (%s) from %s", __FUNCTION__,
                         icmp->icmp_type, icmp->icmp_code,
                         getIcmpType(icmp->icmp_type,icmp->icmp_code),
                         xinet_ntoa (sp->sin_addr.s_addr, 
                                     doResolv & RES_RUNTIME));
/*
	we're only interested in 'unreachable' messages
	if we detect 'port unreachable', this port is stored in the reply
	structure too, otherwise sin_port will be zero.
*/
    sp->sin_port = 0;

    if ((icmp->icmp_type == ICMP_UNREACH) &&
        (icmp->icmp_code == ICMP_PORT_UNREACH)) {
        if (recvLen >= (ipHeadLen + ICMP_MINLEN + (int)sizeof(struct ip))) {
            ip2Head      = (struct ip *)&icmp->icmp_ip;
            ip2HeadLen   = ip2Head->ip_hl << 2;
            if (recvLen >= (ipHeadLen + ICMP_MINLEN + ip2HeadLen +
                            (int)sizeof(struct udphdr))) {
                udpHead             = (struct udphdr *)(((char *)ip2Head)+
                                                        ip2HeadLen);
                sp->sin_addr        = ip2Head->ip_dst;
                sp->sin_port        = udpHead->UDP_DPORT;
                return 1;
            }
        }
    }
    return 0;
}
